//
//  FDAppDelegate.m
//  Durango
//
//  Created by Josh Buhler on 12/20/12.
//  Copyright (c) 2012 Josh Buhler. All rights reserved.
//

#import "FDAppDelegate.h"
#import "AFNetworking.h"
#import "JSONKit.h"
#import "XBGame.h"
#import "XBAchievement.h"
#import "XBGamer.h"
#import "PlayerAchievement.h"

@interface FDAppDelegate()
{
    BOOL    _refreshAllAchievements;
    
    NSOperationQueue    *_achievementQueue;
    int _queuedAchievements;
    int _completedAchievements;
}

@end

@implementation FDAppDelegate

@synthesize persistentStoreCoordinator = _persistentStoreCoordinator;
@synthesize managedObjectModel = _managedObjectModel;
@synthesize managedObjectContext = _managedObjectContext;

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    // Insert code here to initialize your application
}

// Returns the directory the application uses to store the Core Data store file. This code uses a directory named "com.fogodev.Durango" in the user's Application Support directory.
- (NSURL *)applicationFilesDirectory
{
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSURL *appSupportURL = [[fileManager URLsForDirectory:NSApplicationSupportDirectory inDomains:NSUserDomainMask] lastObject];
    return [appSupportURL URLByAppendingPathComponent:@"com.fogodev.Durango"];
}

// Creates if necessary and returns the managed object model for the application.
- (NSManagedObjectModel *)managedObjectModel
{
    if (_managedObjectModel) {
        return _managedObjectModel;
    }
	
    NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"Durango" withExtension:@"momd"];
    _managedObjectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];
    return _managedObjectModel;
}

// Returns the persistent store coordinator for the application. This implementation creates and return a coordinator, having added the store for the application to it. (The directory for the store is created, if necessary.)
- (NSPersistentStoreCoordinator *)persistentStoreCoordinator
{
    if (_persistentStoreCoordinator) {
        return _persistentStoreCoordinator;
    }
    
    NSManagedObjectModel *mom = [self managedObjectModel];
    if (!mom) {
        NSLog(@"%@:%@ No model to generate a store from", [self class], NSStringFromSelector(_cmd));
        return nil;
    }
    
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSURL *applicationFilesDirectory = [self applicationFilesDirectory];
    NSError *error = nil;
    
    NSDictionary *properties = [applicationFilesDirectory resourceValuesForKeys:@[NSURLIsDirectoryKey] error:&error];
    
    if (!properties) {
        BOOL ok = NO;
        if ([error code] == NSFileReadNoSuchFileError) {
            ok = [fileManager createDirectoryAtPath:[applicationFilesDirectory path] withIntermediateDirectories:YES attributes:nil error:&error];
        }
        if (!ok) {
            [[NSApplication sharedApplication] presentError:error];
            return nil;
        }
    } else {
        if (![properties[NSURLIsDirectoryKey] boolValue]) {
            // Customize and localize this error.
            NSString *failureDescription = [NSString stringWithFormat:@"Expected a folder to store application data, found a file (%@).", [applicationFilesDirectory path]];
            
            NSMutableDictionary *dict = [NSMutableDictionary dictionary];
            [dict setValue:failureDescription forKey:NSLocalizedDescriptionKey];
            error = [NSError errorWithDomain:@"YOUR_ERROR_DOMAIN" code:101 userInfo:dict];
            
            [[NSApplication sharedApplication] presentError:error];
            return nil;
        }
    }
    
    NSURL *url = [applicationFilesDirectory URLByAppendingPathComponent:@"Durango.storedata"];
    NSPersistentStoreCoordinator *coordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:mom];
    if (![coordinator addPersistentStoreWithType:NSXMLStoreType configuration:nil URL:url options:nil error:&error]) {
        [[NSApplication sharedApplication] presentError:error];
        return nil;
    }
    _persistentStoreCoordinator = coordinator;
    
    return _persistentStoreCoordinator;
}

// Returns the managed object context for the application (which is already bound to the persistent store coordinator for the application.) 
- (NSManagedObjectContext *)managedObjectContext
{
    if (_managedObjectContext) {
        return _managedObjectContext;
    }
    
    NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
    if (!coordinator) {
        NSMutableDictionary *dict = [NSMutableDictionary dictionary];
        [dict setValue:@"Failed to initialize the store" forKey:NSLocalizedDescriptionKey];
        [dict setValue:@"There was an error building up the data file." forKey:NSLocalizedFailureReasonErrorKey];
        NSError *error = [NSError errorWithDomain:@"YOUR_ERROR_DOMAIN" code:9999 userInfo:dict];
        [[NSApplication sharedApplication] presentError:error];
        return nil;
    }
    _managedObjectContext = [[NSManagedObjectContext alloc] init];
    [_managedObjectContext setPersistentStoreCoordinator:coordinator];

    return _managedObjectContext;
}

// Returns the NSUndoManager for the application. In this case, the manager returned is that of the managed object context for the application.
- (NSUndoManager *)windowWillReturnUndoManager:(NSWindow *)window
{
    return [[self managedObjectContext] undoManager];
}

// Performs the save action for the application, which is to send the save: message to the application's managed object context. Any encountered errors are presented to the user.
- (IBAction)saveAction:(id)sender
{
    NSError *error = nil;
    
    if (![[self managedObjectContext] commitEditing]) {
        NSLog(@"%@:%@ unable to commit editing before saving", [self class], NSStringFromSelector(_cmd));
    }
    
    if (![[self managedObjectContext] save:&error]) {
        [[NSApplication sharedApplication] presentError:error];
    }
}

- (IBAction)onRefreshClick:(id)sender {
    [self requestGamertagInfo:[self.gamertagTxt stringValue]];
}

- (IBAction)onGamesClick:(id)sender
{
    [self getGamesForGamer:[self.gamertagTxt stringValue]];    
}

- (IBAction)onAchievementsClick:(id)sender
{    
    //
}

- (NSString *)getEncodedGamertag
{
    return [[self.gamertagTxt stringValue] stringByAddingPercentEscapesUsingEncoding:NSASCIIStringEncoding];
}



- (void) requestGamertagInfo:(NSString *)gamertag
{
    NSLog(@"Requesting gamertag info for %@", gamertag);
    
    NSString *urlString = [NSString stringWithFormat:@"http://www.xboxleaders.com/api/profile.json?gamertag=%@",
                           [self getEncodedGamertag]];
    NSURL *url = [NSURL URLWithString:urlString];
    NSURLRequest *request = [NSURLRequest requestWithURL:url];
    
    AFJSONRequestOperation *op = [AFJSONRequestOperation JSONRequestOperationWithRequest:request
                                                                                 success:^(NSURLRequest *request, NSHTTPURLResponse *response, id JSON) {
                                                                                     NSLog(@"success: %@", JSON);
                                                                                 } failure:^(NSURLRequest *request, NSHTTPURLResponse *response, NSError *error, id JSON) {
                                                                                     NSLog(@"failed");
                                                                                 }];
    [op start];
}

- (void) getGamesForGamer:(NSString *)gamertag
{
    
    NSLog(@"Requesting games info for %@", gamertag);
    
    [self.achievementProgress setIndeterminate:YES];
    [self.achievementProgress startAnimation:nil];
    
    NSString *urlString = [NSString stringWithFormat:@"http://www.xboxleaders.com/api/games.json?gamertag=%@",
                           [self getEncodedGamertag]];
    NSURL *url = [NSURL URLWithString:urlString];
    NSURLRequest *request = [NSURLRequest requestWithURL:url];
    
    if (_achievementQueue != nil)
        [_achievementQueue cancelAllOperations];
    
    AFJSONRequestOperation *op = [AFJSONRequestOperation JSONRequestOperationWithRequest:request
                                                                                 success:^(NSURLRequest *request, NSHTTPURLResponse *response, id JSON) {
                                                                                     //NSLog(@"success: %@", JSON);
                                                                                     [self parseGamesResponse:JSON];
                                                                                     
                                                                                 } failure:^(NSURLRequest *request, NSHTTPURLResponse *response, NSError *error, id JSON) {
                                                                                     NSLog(@"failed");
                                                                                 }];
    [op start];
}

- (void) parseGamesResponse:(id)JSON
{
    NSDictionary *data = [JSON objectForKey:@"Data"];
    
    if ([(NSArray *)data count] <= 0)
    {
        NSLog(@"No game data");
        return;
    }
    
    // get the gamer from the local database
    XBGamer *gamer = nil;
    
    NSString *gamertag = [data objectForKey:@"Gamertag"];
    NSFetchRequest *fetch = [[NSFetchRequest alloc] init];
    [fetch setEntity:[NSEntityDescription entityForName:@"XBGamer"
                                 inManagedObjectContext:self.managedObjectContext]];
    [fetch setPredicate:[NSPredicate predicateWithFormat:@"gamertag == %@", gamertag]];
    
    NSError *gamertagFetchError = nil;
    NSArray *tagResults = [self.managedObjectContext executeFetchRequest:fetch
                                                                   error:&gamertagFetchError];
    if ([tagResults count] > 0)
    {
        gamer = (XBGamer *)[tagResults objectAtIndex:0];
        NSLog(@"    found player: %@ with %ld games", gamertag, [gamer.games count]);
    }
    
    if (gamer == nil)
    {
        // create the gamer
        NSLog(@"    creating player record");
        gamer = (XBGamer *)[NSEntityDescription insertNewObjectForEntityForName:@"XBGamer"
                                                         inManagedObjectContext:self.managedObjectContext];
        [gamer setGamertag:gamertag];
        [gamer setGamerpic:[data objectForKey:@"Gamerpic"]];
    }
    
    NSArray *games = [data objectForKey:@"PlayedGames"];
    NSLog(@"    total json games: %ld", [games count]);
    
    _queuedAchievements = 0;
    
    NSMutableArray *gameList = [[NSMutableArray alloc] init];
    
    for (NSDictionary *cGame in games)
    {
        NSLog(@"game: %@", [cGame objectForKey:@"Title"]);
        
        XBGame *xGame = nil;
        
        NSFetchRequest *fetch = [[NSFetchRequest alloc] init];
        [fetch setEntity:[NSEntityDescription entityForName:@"XBGame"
                                     inManagedObjectContext:self.managedObjectContext]];
        [fetch setPredicate:[NSPredicate predicateWithFormat:@"gameID == %@", [(NSNumber *)[cGame objectForKey:@"Id"] stringValue]]];
        
        NSError *fetchError = nil;
        NSMutableArray *results = [[self.managedObjectContext executeFetchRequest:fetch
                                                                            error:&fetchError] mutableCopy];
        if ([results count] > 0)
        {
            xGame = (XBGame *)[results objectAtIndex:0];
            NSLog(@"    found game");
        }
        
        if (xGame == nil)
        {
            NSLog(@"    creating game");
            xGame = (XBGame *)[NSEntityDescription insertNewObjectForEntityForName:@"XBGame"
                                                            inManagedObjectContext:self.managedObjectContext];
        }
        
        [xGame setBoxArt:[cGame objectForKey:@"BoxArt"]];
        [xGame setEarnedAchievements:[cGame objectForKey:@"EarnedAchievements"]];
        [xGame setEarnedGamerScore:[cGame objectForKey:@"EarnedGamerScore"]];
        [xGame setGameID:[(NSNumber *)[cGame objectForKey:@"Id"] stringValue]];
        [xGame setLargeBoxArt:[cGame objectForKey:@"LargeBoxArt"]];
        [xGame setLastPlayed:[NSDate dateWithTimeIntervalSince1970:[(NSNumber *)[cGame objectForKey:@"LastPlayed"] intValue]]];
        [xGame setPercentageCompleted:[cGame objectForKey:@"PercentageCompleted"]];
        [xGame setPossibleAchievements:[cGame objectForKey:@"PossibleAchievements"]];
        [xGame setPossibleGamerScore:[cGame objectForKey:@"PossibleGamerScore"]];
        [xGame setTitle:[cGame objectForKey:@"Title"]];
        [xGame setUrl:[cGame objectForKey:@"Url"]];
        
        [gamer addGamesObject:xGame];
        
        // should we also load the achievements for this?
        if (self.refreshAllBox.state == NSOnState)
        {
            // queue achievement operation
            NSLog(@"    queueing achievement refresh");
            
            [gameList addObject:xGame];
        }
    }
    
    NSError *error = nil;
    if (![self.managedObjectContext save:&error])
    {
        NSLog(@"There was an error saving the context");
    }
    
    [self.achievementProgress stopAnimation:nil];
    [self.achievementProgress setIndeterminate:NO];
    
    if (self.refreshAllBox.state == NSOnState)
    {
        // now that the context is saved, queue up the achievements        
        for (XBGame *cGame in gameList)
            [self queueAchievementRefresh:cGame];
    }
}

- (void) queueAchievementRefresh:(XBGame *)game
{
    _queuedAchievements++;
    
    if (_achievementQueue == nil)
    {
        _achievementQueue = [[NSOperationQueue alloc] init];
        [_achievementQueue setMaxConcurrentOperationCount:5];
    }
    
    self.achievementProgress.maxValue = _queuedAchievements;
    
    NSString *urlString = [NSString stringWithFormat:@"http://www.xboxleaders.com/api/achievements.json?gamertag=%@&titleid=%@",
                           [self getEncodedGamertag], game.gameID];
    NSURL *url = [NSURL URLWithString:urlString];
    NSURLRequest *request = [NSURLRequest requestWithURL:url];
    
    AFJSONRequestOperation *op = [AFJSONRequestOperation JSONRequestOperationWithRequest:request
                                                                                 success:^(NSURLRequest *request, NSHTTPURLResponse *response, id JSON) {
                                                                                     //NSLog(@"success: %@", JSON);
                                                                                     [self parseAchievementResponse:JSON];
                                                                                     
                                                                                 } failure:^(NSURLRequest *request, NSHTTPURLResponse *response, NSError *error, id JSON) {
                                                                                     NSLog(@"failed");
                                                                                 }];
    
    [_achievementQueue addOperation:op];
}

- (void) parseAchievementResponse:(id)JSON
{
    _completedAchievements++;
    
    //NSLog(@"completed: %d - %@", _completedAchievements, JSON);
    
    if (_achievementQueue != nil)
    {
        dispatch_async(dispatch_get_main_queue(), ^{
            [self.achievementProgress incrementBy:1.0f];
        });        
    }
    
    NSDictionary *data = [JSON objectForKey:@"Data"];
    
    if ([(NSArray *)data count] <= 0)
    {
        NSLog(@"No game data");
        return;
    }
    
    // get the gamer from the local database
    XBGamer *gamer = nil;
    
    NSString *gamertag = [self.gamertagTxt stringValue];
    NSFetchRequest *fetch = [[NSFetchRequest alloc] init];
    [fetch setEntity:[NSEntityDescription entityForName:@"XBGamer"
                                 inManagedObjectContext:self.managedObjectContext]];
    [fetch setPredicate:[NSPredicate predicateWithFormat:@"gamertag == %@", gamertag]];
    
    NSError *gamertagFetchError = nil;
    NSArray *tagResults = [self.managedObjectContext executeFetchRequest:fetch
                                                                   error:&gamertagFetchError];
    if ([tagResults count] > 0)
    {
        gamer = (XBGamer *)[tagResults objectAtIndex:0];
    }
    
    if (gamer == nil)
    {
        // create the gamer
        NSLog(@"    *** gamer not found");
        return;
    }
    
    // get the game
    // get the gamer from the local database
    XBGame *game = nil;
    
    NSString *gameID = [[data objectForKey:@"Id"] stringValue];
    NSFetchRequest *gameFetch = [[NSFetchRequest alloc] init];
    [gameFetch setEntity:[NSEntityDescription entityForName:@"XBGame"
                                 inManagedObjectContext:self.managedObjectContext]];
    [gameFetch setPredicate:[NSPredicate predicateWithFormat:@"gameID == %@", gameID]];
    
    NSError *gameFetchError = nil;
    NSArray *gameResults = [self.managedObjectContext executeFetchRequest:gameFetch
                                                                   error:&gameFetchError];
    if ([gameResults count] > 0)
    {
        game = (XBGame *)[gameResults objectAtIndex:0];
        NSLog(@"    found game: %@", game.title);
    }
    
    if (game == nil)
    {
        // create the gamer
        NSLog(@"    *** game not found");
        return;
    }
    
    // now process the achievements    
    NSArray *cheevs = [data objectForKey:@"Achievements"];
    NSLog(@"    total json achievements: %ld", [cheevs count]);
    
    for (NSDictionary *cCheev in cheevs)
    {
        NSLog(@"Achievement: %@", [cCheev objectForKey:@"Title"]);
        
        XBAchievement *xCheev = nil;
        
        NSFetchRequest *fetch = [[NSFetchRequest alloc] init];
        [fetch setEntity:[NSEntityDescription entityForName:@"XBAchievement"
                                     inManagedObjectContext:self.managedObjectContext]];
        [fetch setPredicate:[NSPredicate predicateWithFormat:@"game == %@ AND achievementID == %@", game, [(NSNumber *)[cCheev objectForKey:@"Id"] stringValue]]];
        
        NSError *fetchError = nil;
        NSMutableArray *results = [[self.managedObjectContext executeFetchRequest:fetch
                                                                            error:&fetchError] mutableCopy];
        if ([results count] > 0)
        {
            xCheev = (XBAchievement *)[results objectAtIndex:0];
            NSLog(@"    found achievement");
        }
        
        if (xCheev == nil)
        {
            NSLog(@"    creating achievement");
            xCheev = (XBAchievement *)[NSEntityDescription insertNewObjectForEntityForName:@"XBAchievement"
                                                            inManagedObjectContext:self.managedObjectContext];
        }
        
        [xCheev setAchievementID:[cCheev objectForKey:@"Id"]];
        [xCheev setTileURL:[cCheev objectForKey:@"TileUrl"]];
        [xCheev setTitle:[cCheev objectForKey:@"Title"]];
        [xCheev setDesc:[cCheev objectForKey:@"Description"]];
        [xCheev setGamerScore:[cCheev objectForKey:@"Id"]];
        
        NSNumber *isSecret = [NSNumber numberWithBool:[[cCheev objectForKey:@"IsSecret"] isEqualToString:@"yes"]];
        [xCheev setIsSecret:isSecret];
        
        [game addAchievementsObject:xCheev];
        
        // link it to the player
        PlayerAchievement *pCheev = nil;
        
        NSFetchRequest *pFetch = [[NSFetchRequest alloc] init];
        [pFetch setEntity:[NSEntityDescription entityForName:@"PlayerAchievement"
                                     inManagedObjectContext:self.managedObjectContext]];
        [pFetch setPredicate:[NSPredicate predicateWithFormat:@"player == %@ AND achievement == %@", gamer, xCheev]];
        
        NSError *playerFetchError = nil;
        NSMutableArray *playerResults = [[self.managedObjectContext executeFetchRequest:pFetch
                                                                            error:&playerFetchError] mutableCopy];
        if ([playerResults count] > 0)
        {
            pCheev = (PlayerAchievement *)[playerResults objectAtIndex:0];
            NSLog(@"    found PlayerAchievement");
        }
        
        if (pCheev == nil)
        {
            NSLog(@"    creating PlayerAchievement");
            pCheev = (PlayerAchievement *)[NSEntityDescription insertNewObjectForEntityForName:@"PlayerAchievement"
                                                                    inManagedObjectContext:self.managedObjectContext];
        }
        
        [pCheev setPlayer:gamer];
        [pCheev setAchievement:xCheev];
        
        [pCheev setDateEarned:[NSDate dateWithTimeIntervalSince1970:[(NSNumber *)[data objectForKey:@"DateEarned"] intValue]]];
        
        NSNumber *unlocked = [NSNumber numberWithBool:[[cCheev objectForKey:@"Unlocked"] isEqualToString:@"yes"]];
        [pCheev setUnlocked:unlocked];
        
        NSNumber *offline = [NSNumber numberWithBool:[[cCheev objectForKey:@"EarnedOffline"] isEqualToString:@"yes"]];
        [pCheev setEarnedOffline:offline];
        
    }
    
    NSError *error = nil;
    if (![self.managedObjectContext save:&error])
    {
        NSLog(@"There was an error saving the context");
    }
}



- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
    // Save changes in the application's managed object context before the application terminates.
    
    if (!_managedObjectContext) {
        return NSTerminateNow;
    }
    
    if (![[self managedObjectContext] commitEditing]) {
        NSLog(@"%@:%@ unable to commit editing to terminate", [self class], NSStringFromSelector(_cmd));
        return NSTerminateCancel;
    }
    
    if (![[self managedObjectContext] hasChanges]) {
        return NSTerminateNow;
    }
    
    NSError *error = nil;
    if (![[self managedObjectContext] save:&error]) {

        // Customize this code block to include application-specific recovery steps.              
        BOOL result = [sender presentError:error];
        if (result) {
            return NSTerminateCancel;
        }

        NSString *question = NSLocalizedString(@"Could not save changes while quitting. Quit anyway?", @"Quit without saves error question message");
        NSString *info = NSLocalizedString(@"Quitting now will lose any changes you have made since the last successful save", @"Quit without saves error question info");
        NSString *quitButton = NSLocalizedString(@"Quit anyway", @"Quit anyway button title");
        NSString *cancelButton = NSLocalizedString(@"Cancel", @"Cancel button title");
        NSAlert *alert = [[NSAlert alloc] init];
        [alert setMessageText:question];
        [alert setInformativeText:info];
        [alert addButtonWithTitle:quitButton];
        [alert addButtonWithTitle:cancelButton];

        NSInteger answer = [alert runModal];
        
        if (answer == NSAlertAlternateReturn) {
            return NSTerminateCancel;
        }
    }

    return NSTerminateNow;
}

@end
